@@ -3,14 +3,14 @@
  */
 
 /* 
- * Copyright (C) 1991-2004 the Free Software Foundation, Inc.
+ * Copyright (C) 1991-2010 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
  * 
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  * 
  * GAWK is distributed in the hope that it will be useful,
@@ -20,7 +20,7 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
  */
 
 #include "awk.h"
@@ -42,10 +42,11 @@ make_regexp(const char *s, size_t len, int ignorecase, int dfa)
 	static short first = TRUE;
 	static short no_dfa = FALSE;
 	int has_anchor = FALSE;
-#ifdef MBS_SUPPORT
-	/* The number of bytes in the current multbyte character.
+
+	/* The number of bytes in the current multibyte character.
 	   It is 0, when the current character is a singlebyte character.  */
 	size_t is_multibyte = 0;
+#ifdef MBS_SUPPORT
 	mbstate_t mbs;
 
 	if (gawk_mb_cur_max > 1)
@@ -69,7 +70,7 @@ make_regexp(const char *s, size_t len, int ignorecase, int dfa)
 
 	while (src < end) {
 #ifdef MBS_SUPPORT
-		if (gawk_mb_cur_max > 1 && !is_multibyte) {
+		if (gawk_mb_cur_max > 1 && ! is_multibyte) {
 			/* The previous byte is a singlebyte character, or last byte
 			   of a multibyte character.  We check the next character.  */
 			is_multibyte = mbrlen(src, end - src, &mbs);
@@ -81,12 +82,9 @@ make_regexp(const char *s, size_t len, int ignorecase, int dfa)
 		}
 #endif
 
-		if (
-#ifdef MBS_SUPPORT
 		/* We skip multibyte character, since it must not be a special
 		   character.  */
-		    (gawk_mb_cur_max == 1 || ! is_multibyte) &&
-#endif
+		if ((gawk_mb_cur_max == 1 || ! is_multibyte) &&
 		    (*src == '\\')) {
 			c = *++src;
 			switch (c) {
@@ -114,7 +112,7 @@ make_regexp(const char *s, size_t len, int ignorecase, int dfa)
 				 * literally in re's, so escape regexp
 				 * metacharacters.
 				 */
-				if (do_traditional && ! do_posix && (ISDIGIT(c) || c == 'x')
+				if (do_traditional && ! do_posix && (isdigit(c) || c == 'x')
 				    && strchr("()|*+?.^$\\[]", c2) != NULL)
 					*dest++ = '\\';
 				*dest++ = (char) c2;
@@ -145,15 +143,14 @@ make_regexp(const char *s, size_t len, int ignorecase, int dfa)
 				has_anchor = TRUE;
 			*dest++ = *src++;	/* not '\\' */
 		}
-#ifdef MBS_SUPPORT
 		if (gawk_mb_cur_max > 1 && is_multibyte)
 			is_multibyte--;
-#endif
 	} /* while */
 
 	*dest = '\0' ;	/* Only necessary if we print dest ? */
 	emalloc(rp, Regexp *, sizeof(*rp), "make_regexp");
 	memset((char *) rp, 0, sizeof(*rp));
+	rp->dfareg = dfaalloc();
 	rp->pat.allocated = 0;	/* regex will allocate the buffer */
 	emalloc(rp->pat.fastmap, char *, 256, "make_regexp");
 
@@ -171,25 +168,19 @@ make_regexp(const char *s, size_t len, int ignorecase, int dfa)
 	 * also think it's probably better for portability.  See the
 	 * discussion by the definition of casetable[] in eval.c.
 	 */
-#ifdef MBS_SUPPORT
+
 	if (ignorecase) {
 		if (gawk_mb_cur_max > 1) {
 			syn |= RE_ICASE;
 			rp->pat.translate = NULL;
 		} else {
 			syn &= ~RE_ICASE;
-			rp->pat.translate = (char *) casetable;
+			rp->pat.translate = (RE_TRANSLATE_TYPE) casetable;
 		}
 	} else {
 		rp->pat.translate = NULL;
 		syn &= ~RE_ICASE;
 	}
-#else /* ! MBS_SUPPORT */
-	if (ignorecase)
-		rp->pat.translate = (char *) casetable;
-	else
-		rp->pat.translate = NULL;
-#endif /* ! MBS_SUPPORT */
 
 	dfasyntax(syn | (ignorecase ? RE_ICASE : 0), ignorecase ? TRUE : FALSE, '\n');
 	re_set_syntax(syn);
@@ -201,7 +192,7 @@ make_regexp(const char *s, size_t len, int ignorecase, int dfa)
 	/* gack. this must be done *after* re_compile_pattern */
 	rp->pat.newline_anchor = FALSE; /* don't get \n in middle of string */
 	if (dfa && ! no_dfa) {
-		dfacomp(temp, len, &(rp->dfareg), TRUE);
+		dfacomp(temp, len, rp->dfareg, TRUE);
 		rp->dfa = TRUE;
 	} else
 		rp->dfa = FALSE;
@@ -215,16 +206,38 @@ make_regexp(const char *s, size_t len, int ignorecase, int dfa)
 
 int
 research(Regexp *rp, register char *str, int start,
-	register size_t len, int need_start)
+	register size_t len, int flags)
 {
 	const char *ret = str;
 	int try_backref;
+	int need_start;
+	int no_bol;
+	int res;
+
+	need_start = ((flags & RE_NEED_START) != 0);
+	no_bol = ((flags & RE_NO_BOL) != 0);
+
+	if (no_bol)
+		rp->pat.not_bol = 1;
 
 	/*
 	 * Always do dfa search if can; if it fails, then even if
 	 * need_start is true, we won't bother with the regex search.
+	 *
+	 * The dfa matcher doesn't have a no_bol flag, so don't bother
+	 * trying it in that case.
+	 *
+	 * 4/2007: Grrrr.  The dfa matcher has bugs in certain multibyte
+	 * cases that are just too deeply buried to ferret out. Don't
+	 * let this kill us if we need_start.  (This may be too narrowly
+	 * focused, perhaps we should relegate the DFA matcher to the
+	 * single byte case all the time. OTOH, the speed difference
+	 * between the matchers in non-trivial... Sigh.)
+	 *
+	 * 7/2008: Simplify: skip dfa matcher if need_start. The above
+	 * problems are too much to deal with.
 	 */
-	if (rp->dfa) {
+	if (rp->dfa && ! no_bol && ! need_start) {
 		char save;
 		int count = 0;
  		/*
@@ -232,7 +245,7 @@ research(Regexp *rp, register char *str, int start,
 		 * text.  So we just save and restore the character.
 		 */
 		save = str[start+len];
-		ret = dfaexec(&(rp->dfareg), str+start, str+start+len, TRUE,
+		ret = dfaexec(rp->dfareg, str+start, str+start+len, TRUE,
 					&count, &try_backref);
 		str[start+len] = save;
 	}
@@ -243,14 +256,15 @@ research(Regexp *rp, register char *str, int start,
 			 * Passing NULL as last arg speeds up search for cases
 			 * where we don't need the start/end info.
 			 */
-			int res = re_search(&(rp->pat), str, start+len,
+			res = re_search(&(rp->pat), str, start+len,
 				start, len, need_start ? &(rp->regs) : NULL);
-
-			return res;
 		} else
-			return 1;
+			res = 1;
 	} else
-		return -1;
+		res = -1;
+
+	rp->pat.not_bol = 0;
+	return res;
 }
 
 /* refree --- free up the dynamic memory used by a compiled regexp */
@@ -271,8 +285,10 @@ refree(Regexp *rp)
 		free(rp->regs.start);
 	if (rp->regs.end)
 		free(rp->regs.end);
-	if (rp->dfa)
-		dfafree(&(rp->dfareg));
+	if (rp->dfa) {
+		dfafree(rp->dfareg);
+		free(rp->dfareg);
+	}
 	free(rp);
 }
  
@@ -292,7 +308,7 @@ re_update(NODE *t)
 	NODE *t1;
 
 	if ((t->re_flags & CASE) == IGNORECASE) {
-		if ((t->re_flags & CONST) != 0) {
+		if ((t->re_flags & CONSTANT) != 0) {
 			assert(t->type == Node_regex);
 			return t->re_reg;
 		}
@@ -386,7 +402,7 @@ reisstring(const char *text, size_t len, Regexp *re, const char *buf)
 	/* make accessable to gdb */
 	matched = &buf[RESTART(re, buf)];
 
-	res = STREQN(text, matched, len);
+	res = (memcmp(text, matched, len) == 0);
 
 	return res;
 }
@@ -434,8 +450,14 @@ reflags2str(int flagval)
 		{ RE_DEBUG, "RE_DEBUG" },
 		{ RE_INVALID_INTERVAL_ORD, "RE_INVALID_INTERVAL_ORD" },
 		{ RE_ICASE, "RE_ICASE" },
+		{ RE_CARET_ANCHORS_HERE, "RE_CARET_ANCHORS_HERE" },
+		{ RE_CONTEXT_INVALID_DUP, "RE_CONTEXT_INVALID_DUP" },
+		{ RE_NO_SUB, "RE_NO_SUB" },
 		{ 0,	NULL },
 	};
 
+	if (flagval == RE_SYNTAX_EMACS) /* == 0 */
+		return "RE_SYNTAX_EMACS";
+
 	return genflags2str(flagval, values);
 }
